import torch 

from .adlib import SVD 
svd = SVD.apply
from .ctmrg import CTMRG
from .vumps import vumps_run, vumps_calc  

def projection(T, epsilon=1E-3):

    D = T.shape[0] # double layer bond dimension

    M = T.view(D, -1)
    M = M@M.t()
    
    U, S, _ = svd(M)
    
    #S = S/S.max()
    #chi = (S>epsilon).sum().item()
    
    #up to truncation error 
    chi = (torch.cumsum(S, dim=0)/S.sum() <= 1-epsilon).sum().item()

    U = U[:, :chi].view(D, chi)
    #print (S/S.max())
    #print (torch.cumsum(S, dim=0)/S.sum() )
    print ('---->truncated from', D, 'to', chi)

    return torch.einsum('abcd,ai,bj,ck,dl->ijkl', (T, U, U, U, U)), U

def ctmrg(T, chi, maxiter, epsilon):
    
    with torch.no_grad():
        C, E = CTMRG(T, chi, maxiter, epsilon)

    Z1 = torch.einsum('ab,bcd,fd,gha,chij,fjk,lg,mil,mk', (C,E,C,E,T,E,C,E,C))
    Z3 = torch.einsum('ab,bc,cd,da', (C,C,C,C))
    Z2 = torch.einsum('ab,bcd,de,fa,gcf,ge',(C,E,C,C,E,C))
    lnZ = torch.log(Z1.abs()) + torch.log(Z3.abs()) - 2.*torch.log(Z2.abs())

    return lnZ

def vumps(T, chi, maxiter, epsilon):
    with torch.no_grad():
        _, AC, F, C, _ = vumps_run(T, chi, epsilon, maxiter)

    lnZ = torch.log(vumps_calc(T, AC, F, C))
    return lnZ 

